Unit FLAT;

{ FLAT 32bits Memory Unit for TP7 by Paul TOTH <tothpaul@free.fr> }

Interface

function FlatPtr(p:pointer):longint;

{
function  ReadByte(Addr:longint):byte;
procedure WriteByte(Addr:longint; Value:byte);
function  ReadWord(Addr:longint):word;
procedure WriteWord(Addr:longint; Value:word);
function  ReadLong(Addr:longint):longint;
procedure WriteLong(Addr:longint; Value:longint);
}

Function ReadByte(Addr:longint):byte; inline(
 $66/$5B/        { POP EBX         }
 $64/$67/$8A/$03 { MOV AL,FS:[EBX] }
);

Procedure WriteByte(Addr:longint; value:byte); inline(
 $58/            { POP AX          }
 $66/$5B/        { POP EBX         }
 $64/$67/$88/$03 { MOV FS:[EBX],AL }
);

Function ReadWord(Addr:longint):word; inline(
 $66/$5B/        { POP EBX         }
 $64/$67/$8B/$03 { MOV AX,FS:[EBX] }
);

Procedure WriteWord(Addr:longint; value:word); inline(
 $58/            { POP AX          }
 $66/$5B/        { POP EBX         }
 $64/$67/$89/$03 { MOV FS:[EBX],AX }
);

Function ReadLong(Addr:longint):longint; inline(
 $66/$5B/            { POP EBX          }
 $64/$67/$66/$8B/$03/{ MOV EAX,FS:[EBX] }
 $66/$8B/$D0/        { MOV EDX,EAX      }
 $66/$C1/$EA/$10     { SHR EDX,16       }
);

Procedure Writelong(Addr:longint; value:longint); inline(
 $66/$58/            { POP EAX          }
 $66/$5B/            { POP EBX          }
 $64/$67/$66/$89/$03 { MOV FS:[EBX],EAX }
);

procedure Move32(Source, Dest, Count: Longint);

procedure FillChar32(Addr,Count:Longint; Value: Byte);
procedure FillWord32(Addr,Count:Longint; Value: Word);
procedure FillLong32(Addr,Count,Value: Longint);

Implementation

function i386:boolean; assembler;
 asm { Check for NT-flat (only on 386+) }
  pushf
   pushf; pop ax
   xor ah,40h; push ax; popf
   pushf; pop bx
  popf
  mov al,TRUE
  cmp ah,bh
  je @Done
   mov al,FALSE
  @Done:
 end;

function V86:boolean; assembler;
 asm { Check for real mode }
  smsw ax
  test al,1
  mov al,TRUE
  jnz @done
   mov al,FALSE
  @done:
 end;

Type
 TSelector=packed record
  Len     :word;
  BaseLow :word;
  BaseHigh:byte;
  Access  :byte;
  Flags   :byte;
  BaseExt :byte;
 end;

 TDescriptor=packed record
  Size    :word;
  Offset  :longint;
 end;

Var
 GDT:TDescriptor;
 Selectors:array[0..1] of TSelector;

function FlatPtr(p:pointer):longint;
 type
  os=record o,s:word end;
 begin
  with os(p) do FlatPtr:=o+longint(s) shl 4;
 end;

procedure FlatMode; assembler;
 asm
  push ds

   cli

   lgdt GDT

  {mov eax,cr0} db $0F,$20,$C0

   or  al,1
  {mov cr0,eax} db $0F,$22,$C0

   jmp @2; @2:

   mov bx,8
   mov ds,bx
   mov es,bx
  {mov fs,bx}   db $8E,$E3
  {mov gs,bx}   db $8E,$EB

   and al,not 1
  {mov cr0,eax} db $0F,$22,$C0

   xor ax,ax
  {mov fs,ax}   db $8E,$E0

   sti

  pop ds
 end;

(*
function  ReadByte(Addr:longint):byte; assembler;
 asm
  {mov ebx,addr   } db $66; mov bx,word[Addr]
  {mov al,fs:[ebx]} db $64,$67,$8A,$03
 end;

procedure WriteByte(Addr:longint; Value:byte); assembler;
 asm
   mov al,Value
  {mov ebx,addr   } db $66; mov bx,word[Addr]
  {mov fs:[ebx],al} db $64,$67,$88,$03
 end;

function  ReadWord(Addr:longint):word; assembler;
 asm
  {mov ebx,addr   } db $66; mov bx,word[Addr]
  {mov ax,fs:[ebx]} db $64,$67,$8B,$03
 end;

procedure WriteWord(Addr:longint; Value:word); assembler;
 asm
   mov ax,Value
  {mov ebx,addr   } db $66; mov bx,word[Addr]
  {mov fs:[ebx],ax} db $64,$67,$89,$03
 end;

function  ReadLong(Addr:longint):longint; assembler;
 asm
  {mov ebx,addr    } db $66; mov bx,word[Addr]
  {mov eax,fs:[ebx]} db $64,$67,$66,$8B,$03
  {mov edx,eax     } db $66; mov dx,ax
  {shr edx,16      } db $66; shr dx,16
 end;

procedure WriteLong(Addr:longint; Value:Longint); assembler;
 asm
  {mov eax,Value   } db $66; mov ax,word(Value)
  {mov ebx,addr    } db $66; mov bx,word[Addr]
  {mov fs:[ebx],eax} db $64,$67,$66,$89,$03
 end;
*)
procedure Move32(Source, Dest, Count: Longint); assembler;
 asm
  cld
  push ds
  {mov esi,Source} db $66; mov si,word[Source]
  {mov edi,Dest  } db $66; mov di,word[Dest]
  {mov ecx,Count } db $66; mov cx,word[Count]
   xor ax,ax
   mov es,ax
   mov ds,ax
   push cx
  {shr ecx,2     } db $66; shr cx,2
  {rep movsd32   } db $66,$67; rep movsw
   pop cx
   and cx,$0003
  {rep movsb32   } db $67; rep movsb
  pop ds
 end;

procedure FillChar32(Addr,Count: Longint; Value: Byte); assembler;
 asm
  cld
  xor ax,ax
  mov es,ax
 {mov edi,Addr } db $66; mov di,word[Addr]
 {mov ecx,Count} db $66; mov cx,word[Count]
  mov al,[Value]
  mov ah,al
  mov bx,ax
 {shl eax,16   } db $66; shl ax,16
  mov ax,bx
  push cx
 {shr ecx,2    } db $66; shr cx,2
 {rep stosd32  } db $66,$67; rep stosw
  pop cx
  and cx,0003h
 {rep stosb32  } db 67h; rep stosb
 end;

procedure FillWord32(Addr,Count: Longint; Value: Word); assembler;
 asm
  cld
  xor ax,ax
  mov es,ax
 {mov edi,Addr } db $66; mov di,word[Addr]
 {mov ecx,Count} db $66; mov cx,word[Count]
  mov ax,[Value]
  mov bx,ax
 {shl eax,16   } db $66; shl ax,16
  mov ax,bx
  push cx
 {shr ecx,1    } db $66; shr cx,1
 {rep stosd32  } db $66,$67; rep stosw
  pop cx
  and cx,0001h
 {rep stosb32  } db 67h; rep stosw
 end;

procedure FillLong32(Addr,Count,Value: Longint); assembler;
 asm
  cld
  xor ax,ax
  mov es,ax
 {mov edi,Addr } db $66; mov di,word[Addr]
 {mov ecx,Count} db $66; mov cx,word[Count]
 {mov eax,Value} db $66; mov ax,word[Value]
 {rep stosd32  } db $66,$67; rep stosw
 end;

begin

 if not i386 then begin
  writeln('This program need a i386 or better processor.');
  halt(10);
 end;

 if v86 then begin
  writeln('This program can''t run under V86 (Windows, EMM386...).');
  halt(20);
 end;

 FillChar(Selectors,SizeOf(Selectors),0);

 Selectors[1].Len :=$FFFF; { 4Gb    }
 Selectors[1].Access:=$92; { Write  }
 Selectors[1].Flags :=$CF; { 32bits }

 GDT.Size  :=SizeOf(Selectors);
 GDT.Offset:=FlatPtr(@Selectors);

 FlatMode;

end.