Unit R128;

{ ATI Rage128 Graphic Unit      }
{ tested on an XPert2000 - 32Mo }

{ (C)2000, Paul TOTH <tothpaul@free.fr> }
{ http://tothpaul.free.fr               }

Interface

Uses
 FLAT, { Needed to Read/Write Rage128 memory (above 1Mo) in Real Mode }
 REGDEF;

(*** STANDARD FEATURES ***)

{ Detect a Rage128, Halt on Error }
Procedure Detect;

{ Set Requested Mode, Return Base Memory Offset, Halt on Error }
Function  SetMode(AWidth,AHeight:word; ABPP:byte):longint;
Procedure TextMode;

Procedure FillScreen(Color:integer);

{ Set a Pixel for any Requested Mode }
Const
 BLACK           = 0;
 DARKBLUE        = 1;
 DARKGREEN       = 2;
 DARKCYAN        = 3;
 DARKRED         = 4;
 DARKMAGENTA     = 5;
 BROWN           = 6;
 LIGHTGRAY       = 7;
 DARKGRAY        = 8;
 LIGHTBLUE       = 9;
 LIGHTGREEN      =10;
 LIGHTCYAN       =11;
 LIGHTRED        =12;
 LIGHTMAGENTA    =13;
 YELLOW          =14;
 WHITE           =15;

Function  GetColor(index:integer):longint;
Procedure PutPixel(x,y:word; color:longint);
Function  GetPixel(x,y:word):longint;

Function BaseMemory:longint;
Function BytesPerPixel:byte;
Function BytesPerLine:word;

(*** ADVANCED FEATURES ***)

Procedure InitEngine;
Procedure ResetEngine;
Procedure Rectangle(x,y,w,h,color:integer);
Procedure BitBlt(x,y,w,h,dx,dy:word);

Procedure LoadFont(FileName:string);
Procedure PrintText(x,y:longint; s:string; foreground,background:integer);

Procedure FillRect(x,y,w,h:integer; Pattern0,Pattern1:longint; ForeColor,BackColor:integer);

Procedure LoadCursor(AWidth,AHeight:word; BaseName:string);
Procedure SetCursorColors(ForeColor,BackColor:integer);
Procedure SetCursorPos(x,y:integer);
Procedure ShowCursor;
Procedure HideCursor;

Procedure Line(x1,y1,x2,y2,color:integer);

Procedure DrawBitmap(X,Y:integer; Bitmap:Array of Longint; W,H,ForeColor,BackColor:integer);

Procedure SetBrush(Pattern:array of Longint);
Procedure FillBrush(x,y,w,h,ForeColor,BackColor:integer);

Procedure BrushLine(x1,y1,x2,y2,BrushType,ForeColor,BackColor:integer);

procedure StrechBitBlt(sx,sy,sw,sh,dx,dy,dw,dh:integer);

Procedure TransBlit(x,y,w,h,dx,dy,mode,srccolor,dstcolor:integer);

(*** Concurrent Command Execution (CCE) programming model ***)
{
Procedure InitCEE;
Procedure DoneCEE;
Procedure SendCEE(Buf:array of longint; count:integer);
}

Implementation

Var
{ PCI Information }
 BUS :byte;    { bus # }
 DEV :byte;    { $1002 for ATI          }
 ID  :word;    { $5246 for my XPert2000 }
{ Memory offset }
 BASE:longint; { $E4000000     }
 IO  :longint; { $0000D000     }
 REGS:longint; { $E9000000     }
 RAM :longint; { RAM Size 32Mo }
 ROM :word;    { ROM Seg $C000 }
{ Mode Info }
 Width :longint;
 Height:longint;
 BPP   :byte;  { 8 15 16 24 32 }
 BPPi  :byte;  { 2  3  4  5  6 }
 BPC   :byte;  { 1  2  2  3  4 }
 Pitch :longint;{ WIDTH/8 (*3 for 24BPP) }

{---Util--------------------}
Function Hexa(L:Longint;Size:integer):string;
 const
  Hx:array[0..$F] of char='0123456789ABCDEF';
 var
  s:string;
 begin
  s:='';
  while Size>0 do begin
   s:=Hx[L and $F]+s;
   L:=L shr 4;
   Dec(Size);
  end;
  Hexa:=s;
 end;

{---PCI Related Function---}
Function CheckPCIBus:integer; assembler;
 asm
  mov ax,$B101 { function number to check for PCI BIOS }
  int $1A
  and cx,$FF
  cmp ah,0
  jz @ok
   mov cx,-1
  @ok:
  mov ax,cx
 end;

function GetPCIWord (bus,dev,reg:word):word; assembler;
 asm
  mov ax,bus
  shl ax,8
  mov bx,dev
  shl bx,3
  add bx,ax
  mov ax,$B109 { function number to get a PCI word }
  mov di,reg
  int 1Ah
  mov ax,cx
 end;

function GetPCIDword(bus,dev,reg:word):longint; assembler;
 asm
  mov ax,bus
  shl ax,8
  mov bx,dev
  shl bx,3
  add bx,ax
  mov ax,$B10A { function number to get a PCI dword }
  mov di,reg
  and di,$FC
  int 1Ah
  mov ax,cx
 {shr ecx,16} db 66h; shr cx,16
  mov dx,cx
 end;
{---------------------------}
Procedure FindROM;
 var
  s:string;
  i:integer;
 begin
  s[0]:=#255;
  ROM:=$C000;
  while ROM<$F000 do begin
   if MemW[ROM:0]=$AA55 then begin { Valide ROM }
    move(Mem[ROM:0],s[1],255);
    if pos('761295520',s)>0 then begin { ATI ROM }
     if (pos('R128',s)+pos('M3',s)>0) then exit; { Rage128 ROM }
     for i:=1 to 2 do begin
      move(Mem[ROM:i*128],s[1],255);
      if (pos('R128',s)+pos('M3',s)>0) then exit; { Rage128 ROM }
     end;
    end;
   end;
   inc(ROM,$1000);
  end;
  WriteLn('ROM not found');
  Halt;
 end;

Procedure RegTest;
 var
  v:longint;
  ok:boolean;
 begin
  ok:=false;
  v:=ReadLong(REGS+$10); { BIOS 0 SCRATCH }

 { Test odd bits for readability. }
  WriteLong(REGS+$10,$55555555);
  if ReadLong(REGS+$10)=$55555555 then begin
  { Test even bits for readability. }
   WriteLong(REGS+$10,$AAAAAAAA);
   ok:=(ReadLong(REGS+$10)=$AAAAAAAA);
  end;

  WriteLong(REGS+$10,v);

  if not OK then begin
   WriteLn('Rage 128 scratch register test FAILED');
   Halt;
  end;

 end;


Procedure Detect;
 var
  BusCount:integer;
 begin
 { First, Check PCI }
  write('Looking for PCI...');
  BusCount:=CheckPCIBus;
 { Scan Buses for an ATI Card }
  for bus:=0 to BusCount do begin
   for dev:=0 to 31 do begin
    if GetPCIWord(bus,dev,0)=$1002 then begin { Found ATI Card }
    { Check for a Rage128 card }
     ID:=GetPCIWord(bus,dev,2);
     case ID of
      $4C45, $4C46,
      $5041..$5058,
      $5245, $5246,
      $524B, $524C,
      $5345..$5347,
      $534B..$534D: begin
       WriteLn('1002/',Hexa(ID,4),' OK');
       BASE:=GetPCIdword(bus,dev,$10) and $FC000000;
       IO  :=GetPCIdword(bus,dev,$14) and $FFFFFF00;
       REGS:=GetPCIdword(bus,dev,$18) and $FFFFC000; { (10*1024) }
       WriteLn('Base Memory = ',Hexa(BASE,8));
       WriteLn('I/O  Base   = ',Hexa(IO  ,8));
       WriteLn('Regs Base   = ',Hexa(REGS,8));
       RAM:=ReadLong(REGS+$F8); { MEM SIZE }
       WriteLn('Memory Size = ',RAM div 1024 div 1024,' Mo');
       FindROM;
       WriteLn('ROM Segment = ',Hexa(ROM,4));
       RegTest;
       Exit;
      end;
     end;
    end;
   end;
  end;
 { Card not found }
  WriteLn('no ATI Rage128 card found.');
  Halt;
 end;

Function SetMode(AWidth,AHeight:word; ABPP:byte):longint;
 Const
  Modes:array[0..9,0..1] of word=(
   ( 320, 200),
   ( 320, 240),
   ( 400, 300),
   ( 512, 384),
   ( 640, 350),
   ( 640, 400),
   ( 640, 480),
   ( 800, 600),
   (1024, 768),
   (1280,1024)
  );
  Code:array[0..9] of byte=(
   $E2,$E3,$E5,$E4,$E6,$E1,$12,$6A,$55,$83
  );
  Colors=#8#15#16#24#32;
 var
  m:byte;
 begin
  Width :=AWidth;
  Height:=AHeight;
  BPP   :=ABPP;
  BPPi  :=pos(chr(BPP),Colors)+1;
  if BPPi>1 then
  for m:=0 to 9 do begin
   if (Width=Modes[m,0]) and (Height=Modes[m,1]) then begin
    m:=Code[m];
    asm
     mov ax,$A000
     mov cl,BPPi
     mov ch,m
     int 10h
     mov m,ah
    end;
    if m<>0 then begin
     writeln('Unsupported mode'); Halt;
    end;
    BPC:=(BPP+1) div 8; { 15 -> 2 }
    Pitch:=Width div 8;
    if BPC=3 then Pitch:=3*Pitch;
    case BPC of
     1{    8}:FillChar32(BASE,Width*Height,0);
     2{15/16}:FillWord32(BASE,Width*Height,0);
     3{   24}:FillChar32(BASE,Width*Height*3,0);
     4{   32}:FillLong32(BASE,Width*Height,0);
    end;
    SetMode:=BASE; { Return Base Memory }
    ResetEngine;
    InitEngine;
    exit;
   end;
  end;
  WriteLn('Invalide mode');
  Halt;
 end;

Procedure TextMode;
 begin
  case BPC of
   1:FillChar32(BASE,Width*Height,0);
   2:FillWord32(BASE,Width*Height,0);
   3:FillChar32(BASE,Width*Height*3,0);
   4:FillLong32(BASE,Width*Height,0);
  end;
  ResetEngine;
  asm mov ax,03h; int 10h end;
 end;

Function GetColor(index:integer):longint;
 Const
  Color15:array[0..15] of longint=(
    $00000000,                         { Black         }
    $00000014,                         { Dark Blue     }
    $00000280,                         { Dark Green    }
    $00000294,                         { Dark Cyan     }
    $00005000,                         { Dark Red      }
    $00005014,                         { Dark Magenta  }
    $00005280,                         { Brown         }
    $00005294,                         { Light Gray    }
    $0000294A,                         { Dark Gray     }
    $0000001F,                         { Light Blue    }
    $000003E0,                         { Light Green   }
    $000003FF,                         { Light Cyan    }
    $00007C00,                         { Light Red     }
    $00007C1F,                         { Light Magenta }
    $00007FE0,                         { Yellow        }
    $00007FFF                          { White         }
  );
  Color16:array[0..15] of longint=(
    $00000000,                         { Black         }
    $00000014,                         { Dark Blue     }
    $000004C0,                         { Dark Green    }
    $000004D4,                         { Dark Cyan     }
    $0000A000,                         { Dark Red      }
    $0000A014,                         { Dark Magenta  }
    $0000A4C0,                         { Brown         }
    $0000A4D4,                         { Light Gray    }
    $0000528A,                         { Dark Gray     }
    $0000001F,                         { Light Blue    }
    $000007E0,                         { Light Green   }
    $000007FF,                         { Light Cyan    }
    $0000F800,                         { Light Red     }
    $0000F81F,                         { Light Magenta }
    $0000FFE0,                         { Yellow        }
    $0000FFFF                          { White         }
  );
  ColorRGB:array[0..15] of longint=(
    $00000000,                         { Black         }
    $0000009E,                         { Dark Blue     }
    $00009E00,                         { Dark Green    }
    $00009E9E,                         { Dark Cyan     }
    $009E0000,                         { Dark Red      }
    $009E009E,                         { Dark Magenta  }
    $009E9E00,                         { Brown         }
    $009E9E9E,                         { Light Gray    }
    $00555555,                         { Dark Gray     }
    $000000FF,                         { Light Blue    }
    $0000FF00,                         { Light Green   }
    $0000FFFF,                         { Light Cyan    }
    $00FF0000,                         { Light Red     }
    $00FF00FF,                         { Light Magenta }
    $00FFFF00,                         { Yellow        }
    $00FFFFFF                          { White         }
  );
 begin
  if Index<0 then
   GetColor:=0
  else
   case BPP of
      8: GetColor:=Index;
     15: GetColor:=Color15 [Index and 15];
     16: GetColor:=Color16 [Index and 15];
    else GetColor:=ColorRGB[Index and 15];
   end;
 end;


Procedure FillScreen(Color:integer);
 begin
  Rectangle(0,0,Width,Height,Color);
 end;

Procedure PutPixel(x,y:word; color:longint);
 var
  Offset:longint;
 begin
  Offset:=BASE+BPC*(y*Width+x);
  case BPC of
   1:WriteByte(Offset,Color);
   2:WriteWord(Offset,Color);
   3:begin
      WriteByte(Offset,Color);
      WriteWord(Offset+1,Color shr 8);
     end;
   4:WriteLong(Offset,Color);
  end;
 end;

Function GetPixel(x,y:word):longint;
 var
  Offset:longint;
 begin
  Offset:=BASE+BPC*(y*Width+x);
  case BPC of
   1:GetPixel:=ReadByte(Offset);
   2:GetPixel:=ReadWord(Offset);
   3:GetPixel:=ReadByte(Offset)+Longint(ReadWord(Offset+1)) shl 8;
   4:GetPixel:=ReadLong(Offset);
  end;
 end;

Function BaseMemory:longint;
 begin
  BaseMemory:=BASE;
 end;

Function BytesPerPixel:byte;
 begin
  BytesPerPixel:=BPC;
 end;

Function BytesPerLine:word;
 begin
  BytesPerLine:=BPC*Width;
 end;

(*** ADVANCED FEATURES ***)

Var
 Ticks:word absolute $40:$6C; { use only Low word }

Const
 FIFO_TIMEOUT=50;
 IDLE_TIMEOUT=50;

procedure FlushPixelCache;
 var
  i:integer;
 begin
 { initiate flush }
  WriteLong(REGS+PC_NGUI_CTLSTAT, ReadLong(REGS+PC_NGUI_CTLSTAT) or 255);
 { check for completion but limit looping to 16384 reads }
  for i:=0 to 16384 do
   if (ReadLong(REGS+PC_NGUI_CTLSTAT) and PC_BUSY)<>PC_BUSY then break;
 end;

procedure PLLWrite(Reg:word; Data:longint);
 begin
  WriteByte(REGS+CLOCK_CNTL_INDEX,(Reg and $1F) or PLL_WR_EN);
  WriteLong(REGS+CLOCK_CNTL_DATA,Data);
 end;

function PLLRead(Reg:word):longint;
 begin
  WriteByte(REGS+CLOCK_CNTL_INDEX,(Reg and $1F));
  PLLRead:=ReadLong(REGS+CLOCK_CNTL_DATA);
 end;

procedure ResetEngine;
 var
  save_genresetcntl,
  save_clockcntlindex,
  save_mclkcntl:longint;
 begin
 { Flush the pixel cache }
  FlushPixelCache;

  save_clockcntlindex := ReadLong(REGS+CLOCK_CNTL_INDEX);
  save_mclkcntl       := PLLRead(MCLK_CNTL);

 { we must now force the engine clocks to active before }
 { performing the engine reset.  We must turn them back on later... }
  PLLWrite(MCLK_CNTL, save_mclkcntl or $00030000);

 { save GEN_RESET_CNTL register }
  save_genresetcntl := ReadLong(REGS+GEN_RESET_CNTL);

 { reset by setting bit, add read delay, then clear bit, add read delay }
  WriteLong(REGS+GEN_RESET_CNTL, save_genresetcntl or SOFT_RESET_GUI);
  ReadLong (REGS+GEN_RESET_CNTL);
  WriteLong(REGS+GEN_RESET_CNTL, save_genresetcntl and (not SOFT_RESET_GUI));
  ReadLong (REGS+GEN_RESET_CNTL);

 { restore engine clocks }
  PLLWrite(MCLK_CNTL, save_mclkcntl);

 { restore the two registers we changed }
  WriteLong(REGS+CLOCK_CNTL_INDEX, save_clockcntlindex);
  WriteLong(REGS+GEN_RESET_CNTL, save_genresetcntl);
 end;

Procedure WaitForFifo(Entries:word);
 var
  start:word;
 begin
  start:=ticks;
  while (ReadWord(REGS+GUI_STAT) and $FFF)<Entries do begin
   if (ticks-start)>FIFO_TIMEOUT then begin
    ResetEngine;
    start:=ticks;
   end;
  end;
 end;

Procedure WaitForIdle;
 var
  start:word;
 begin
 { Insure FIFO is empty before waiting for engine idle. }
  WaitForFifo(64);
  start:=Ticks;
  while (ReadLong(REGS+GUI_STAT) and GUI_ACTIVE)<>0 do begin
   if (ticks - start)>IDLE_TIMEOUT then begin
    ResetEngine;
    Start:=Ticks;
   end;
  end;
 { flush the pixel cache }
  FlushPixelCache;
 end;

Procedure InitEngine;
 begin
 { insure 3D is disabled }
  WriteLong(REGS+SCALE_3D_CNTL,0);
 { do an Engine Reset, just in case it's hung right now }
  ResetEngine;
 { setup engine offset registers }
  WaitForFifo(1);
  WriteLong(REGS+DEFAULT_OFFSET,0);
 { setup engine pitch registers }
  if BPP=24 then
   WriteLong(REGS+DEFAULT_PITCH,3*(WIDTH div 8))
  else
   WriteLong(REGS+DEFAULT_PITCH,WIDTH div 8);
 { set scissors to maximum size }
  WaitForFifo(1);
  WriteLong(REGS+DEFAULT_SC_BOTTOM_RIGHT,($1FFF shl 16) or $1FFF);
 { Set the drawing controls registers. }
  WaitForFifo(1);
  WriteLong(REGS+DP_GUI_MASTER_CNTL,
   GMC_SRC_PITCH_OFFSET_DEFAULT or
   GMC_DST_PITCH_OFFSET_DEFAULT or
   GMC_SRC_CLIP_DEFAULT         or
   GMC_DST_CLIP_DEFAULT         or
   GMC_BRUSH_SOLIDCOLOR         or
   (BPPi shl 8)                 or
   GMC_SRC_DSTCOLOR             or
   GMC_BYTE_ORDER_MSB_TO_LSB    or
   GMC_DP_CONVERSION_TEMP_6500  or
   ROP3_PATCOPY                 or
   GMC_DP_SRC_RECT              or
   GMC_3D_FCN_EN_CLR            or
   GMC_DST_CLR_CMP_FCN_CLEAR    or
   GMC_AUX_CLIP_CLEAR           or
   GMC_WRITE_MASK_SET
  );

  WaitForFifo(8);

 { Clear the line drawing registers. }
  WriteLong(REGS+DST_BRES_ERR, 0);
  WriteLong(REGS+DST_BRES_INC, 0);
  WriteLong(REGS+DST_BRES_DEC, 0);

 { set brush color registers }
  WriteLong(REGS+DP_BRUSH_FRGD_CLR, $FFFFFFFF);
  WriteLong(REGS+DP_BRUSH_BKGD_CLR, $00000000);

 { set source color registers }
  WriteLong(REGS+DP_SRC_FRGD_CLR, $FFFFFFFF);
  WriteLong(REGS+DP_SRC_BKGD_CLR, $00000000);

 { default write mask }
  WriteLong(REGS+DP_WRITE_MASK, $FFFFFFFF);

 { Wait for all the writes to be completed before returning }
  WaitForIdle;
 end;

Procedure Rectangle(x,y,w,h,color:integer);
 var
  save_dp_datatype:longint;
 begin
 { Set the drawing controls registers. *** missing in ATI demo :) *** }
  WaitForFifo(1);
  WriteLong(REGS+DP_MIX, ROP3_PATCOPY);

  WaitForFifo(6);
 { Save the current DP_DATATYPE register value }
  save_dp_datatype:=ReadLong(REGS+DP_DATATYPE);

  WriteLong(REGS+DP_DATATYPE, BPPi or BRUSH_SOLIDCOLOR or ROP3_SRCCOPY);
  WriteLong(REGS+DP_BRUSH_FRGD_CLR, GetColor(color));
  WriteLong(REGS+DST_Y_X, longint(y) shl 16+x);
  WriteLong(REGS+DST_WIDTH_HEIGHT, longint(w) shl 16+h);

 { restore the DP_DATATYPE register }
  WriteLong(REGS+DP_DATATYPE, save_dp_datatype);
 end;

Procedure BitBlt(x,y,w,h,dx,dy:word);
 begin
  WaitForFifo(4);
 { Tell the engine where the source data resides. }
  WriteLong(REGS+SRC_Y_X,longint(y) shl 16+x);
 { set the mix to a ROP3_SRCCOPY, and a rectangular source }
  WriteLong(REGS+DP_MIX, ROP3_SRCCOPY or DP_SRC_RECT);
 { Set the drawing direction to left->right, top->bottom }
  WriteLong(REGS+DP_CNTL, DST_X_LEFT_TO_RIGHT or DST_Y_TOP_TO_BOTTOM);
 { make the src pixel = dst pixel, in terms of pixel depth }
  WriteLong(REGS+DP_DATATYPE,ReadLong(REGS+DP_DATATYPE) or SRC_DSTCOLOR);

  WaitForFifo(2);
 { set the destination x and y values }
  WriteLong(REGS+DST_Y_X,longint(dy) shl 16+dx);
 { this is the blt initiator. }
  WriteLong(REGS+DST_HEIGHT_WIDTH,longint(h) shl 16+w);
 end;


Var
 Font:array[0..2495] of byte;

Procedure LoadFont(FileName:string);
 var
  f:file;
 begin
  assign(f,FileName);
  reset(f,1);
  if ioresult<>0 then begin
   TextMode;
   writeln('Error while opening FONT file');
   exit;
  end;
  blockread(f,Font,SizeOf(Font));
  if ioresult<>0 then begin
   TextMode;
   writeln('Error reading FONT file');
   exit;
  end;
  close(f);
 end;

Procedure PrintSpace(x,y:longint);
 var
  i:integer;
 begin
  for i:=0 to 31 do begin
   WaitForFifo(1);
   WriteLong(REGS+BRUSH_DATA0+i*4,0);
  end;
  WriteLong(REGS+DST_Y_X,y shl 16 + x);
  WriteLong(REGS+DST_HEIGHT_WIDTH,16 shl 16+9);
 end;

Const
 RomSize=16;
Type
 TRom8x8=array[#0..#255,0..RomSize-1] of byte;

Var
 rom8x8:^Trom8x8;

Procedure PrintChar(c:char; x,y:longint);
 var
  row,col,offset:longint;
  loc:longint;
  i:integer;
 begin
 (* ROM Font !
  for i:=0 to RomSize-1 do begin
   WaitForFifo(1);
   WriteLong(REGS+BRUSH_DATA0+i*4,rom8x8^[c,i]);
  end;
  for i:=RomSize to 31 do begin
   WaitForFifo(1);
   WriteLong(REGS+BRUSH_DATA0+i*4,0);
  end;
   WaitForFifo(3);
   WriteLong(REGS+BRUSH_Y_X,(32-y mod 32) shl 8 + (32-x mod 32));
   WriteLong(REGS+DST_Y_X,y shl 16 + x);
   WriteLong(REGS+DST_HEIGHT_WIDTH,16 shl 16+9);

  exit;
  *)
  case c of
   'a'..'z':begin row:=0; offset:=97 end;
   'A'..'Z':begin row:=1; offset:=65 end;
   '0'..'9':begin row:=2; offset:=48 end;
   else     begin PrintSpace(x,y); exit end;
  end;
  col:=ord(c)-offset;
 { calculate where in the Font array this letter starts }
  loc:=(26*row*(2*16))+(col*2);
 { since we're using a 32x32 brush, we must load 32 DWORDs into the
   brush data registers. }
  for i:=0 to 31 do begin
   WaitForFifo(1);
   WriteLong(REGS+BRUSH_DATA0+i*4,font[loc]);
   inc(loc,26*2);
  end;

 { now do a 16x9 blt!
   shiftx determines how much we will have to shift the brushes
   to position the pattern at the right edge of the rectangle
   shifty determines how much we will have to shift the brushes
   to position the pattern at the top edge of the rectangle }

 { We will adjust the brushes so that the pattern
   is always starting in the top right of the rectangle }
   WaitForFifo(3);
   WriteLong(REGS+BRUSH_Y_X,(32-y mod 32) shl 8 + (32-x mod 32));
   WriteLong(REGS+DST_Y_X,y shl 16 + x);
   WriteLong(REGS+DST_HEIGHT_WIDTH,16 shl 16+9);
 end;

Procedure PrintText(x,y:longint; s:string; foreground,background:integer);
 var
  save_dp_datatype,
  save_dp_mix,
  save_clr_cmp_cntl,
  save_scale_3d_cntl,
  save_tex_cntl:longint;
  i:integer;
  brush:longint;
 begin
 { save DP_DATATYPE and restore it later }
  WaitForIdle;
  save_dp_datatype   := ReadLong(REGS+DP_DATATYPE);
  save_dp_mix        := ReadLong(REGS+DP_MIX);
  save_clr_cmp_cntl  := ReadLong(REGS+CLR_CMP_CNTL);
  save_scale_3d_cntl := ReadLong(REGS+SCALE_3D_CNTL);
  save_tex_cntl      := ReadLong(REGS+TEX_CNTL);

 { disable any 3D stuff happening. }
  WriteLong(REGS+SCALE_3D_CNTL,0);
  WaitForIdle;

 { clear colour compare cntl }
  WriteLong(REGS+CLR_CMP_CNTL,0);

 { We will use a 32x32 mono pattern, expanding it to the foreground, }
 { and the background may or may not be expanded as well.            }
 { set DP_DATATYPE to satisfy the above conditions.                  }
  if background<0 then
   brush:=BRUSH_32x32MONO_LBKGD
  else
   brush:=BRUSH_32x32MONO;

  WriteLong(REGS+DP_DATATYPE, SRC_MONO or brush or BPPi);
  WriteLong(REGS+DP_MIX, ROP3_PATCOPY or DP_SRC_RECT);
  WriteLong(REGS+DP_BRUSH_FRGD_CLR, GetColor(foreground));
  WriteLong(REGS+DP_BRUSH_BKGD_CLR, GetColor(background));

 { now render the text }
  for i:=1 to length(s) do PrintChar(s[i],x+i*9,y);

  WaitForFifo(4);
  WriteLong(REGS+SCALE_3D_CNTL, save_scale_3d_cntl);
  WriteLong(REGS+DP_DATATYPE, save_dp_datatype);
  WriteLong(REGS+DP_MIX, save_dp_mix);
  WriteLong(REGS+CLR_CMP_CNTL, save_clr_cmp_cntl);
 end;


Procedure FillRect(x,y,w,h:integer; Pattern0,Pattern1:longint; ForeColor,BackColor:integer);
 begin
 { Set the drawing controls registers. *** missing in ATI demo :) *** }
  WaitForFifo(1);
  WriteLong(REGS+DP_MIX, ROP3_PATCOPY);

 { A rectangle is drawn, filled by a colour expanded 8x8 monochrome }
 { fixed pattern via the brush data registers.                      }
 { mono source       = mono 8x8 fixed pattern,                      }
 { foreground source = foreground colour,                           }
 { background source = background colour                            }
  WaitForFifo(3);

 { set DP_DATATYPE to satisfy the above conditions. }
  if BackColor<0 then
   WriteLong(REGS+DP_DATATYPE, BRUSH_8x8MONO_LBKGD or SRC_DSTCOLOR or BPPi)
  else
   WriteLong(REGS+DP_DATATYPE, BRUSH_8x8MONO or SRC_DSTCOLOR or BPPi);

 { This is an 8x8 monochrome pattern of a box with a border. }
  WriteLong(REGS+BRUSH_DATA0, Pattern0);
  WriteLong(REGS+BRUSH_DATA1, Pattern1);

 { foreground colour is used when pattern bit is "1" }
 { background colour is used when pattern bit is "0" }
  WaitForFifo(4);
  WriteLong(REGS+DP_BRUSH_FRGD_CLR, GetColor(ForeColor));
  WriteLong(REGS+DP_BRUSH_BKGD_CLR, GetColor(BackColor));

 { Draw a filled pattern rectangle. }
  WriteLong(REGS+DST_Y_X, longint(y) shl 16+x);
  WriteLong(REGS+DST_HEIGHT_WIDTH, Longint(h) shl 16+w);
 end;


Procedure LoadCursor(AWidth,AHeight:word; BaseName:string);
 var
  f:file;
  Cursor:array[0..1,0..63,0..1] of longint;
  Align :integer;
  x,y   :longint;
  Offset:longint;
 begin
 { Read file }
  assign(f,BaseName+'AND.ICN'); { AND mask }
  reset(f,1);
  blockread(f,Cursor[0],512);
  close(f);
  assign(f,BaseName+'XOR.ICN'); { XOR mask }
  reset(f,1);
  blockread(f,Cursor[1],512);
  close(f);
 { Move Cursor into Video Memory }
  Offset:=Width*Height*BPC+10*(Width*BPC);
  Align:=Offset mod 4;
  if Align<>0 then Inc(Offset,4-Align);
  for y:=0 to 63 do begin
   for x:=0 to 1 do begin
    WriteLong(BASE+Offset+16*y+4*x   ,Cursor[0,y,x]);
    WriteLong(BASE+Offset+16*y+4*x+8 ,Cursor[1,y,x]);
   end;
  end;
 { Check for Double Scan }
  if odd(ReadLong(REGS+CRTC_GEN_CNTL)) then AHeight:=2*AHeight;
  WriteLong(REGS+CUR_HORZ_VERT_OFF,longint(64-AWidth) shl 16+(64-AHeight));
  WriteLong(REGS+CUR_OFFSET,Offset);
 end;

Procedure SetCursorColors(ForeColor,BackColor:integer);
 begin
  WriteLong(REGS+CUR_CLR1, GetColor(ForeColor));
  WriteLong(REGS+CUR_CLR0, GetColor(BackColor));
 end;

Procedure SetCursorPos(x,y:integer);
 begin
  WriteLong(REGS+CUR_HORZ_VERT_POSN, longint(x) shl 16+y);
 end;

Procedure ShowCursor;
 begin
  WriteLong(REGS+CRTC_GEN_CNTL, ReadLong(REGS+CRTC_GEN_CNTL) or $00010000);
 end;

Procedure HideCursor;
 begin
  WriteLong(REGS+CRTC_GEN_CNTL, ReadLong(REGS+CRTC_GEN_CNTL) and (NOT $00010000));
 end;

Procedure Line(x1,y1,x2,y2,color:integer);
 var
  dx,dy:integer;
  flags:longint;
  err,add,sub:longint;
  save_dp_cntl,save_dp_datatype:longint;
 begin
  flags:=0;
  dx:=x2-x1; if dx<0 then dx:=-dx else flags:=          1 shl 31;
  dy:=y2-y1; if dy<0 then dy:=-dy else flags:=flags or (1 shl 15);

 { Calculate Bresenham parameters and draw line. }
  if dx<dy then begin
   flags:=flags or (1 shl 2);
   err:=-dy;
   add:=2*dx;
  end else begin
   err:=-dx;
   add:=2*dy;
  end;

  sub:=2*err; { should be computed by ATI :) }

  WaitForFifo(11);

  save_dp_cntl     := ReadLong(REGS+DP_CNTL);
  save_dp_datatype := ReadLong(REGS+DP_DATATYPE);

 { Set DP_DATATYPE }
  WriteLong(REGS+DP_DATATYPE, (BPPi or BRUSH_SOLIDCOLOR or ROP3_SRCCOPY));
  WriteLong(REGS+DP_BRUSH_FRGD_CLR, GetColor(color));

 { Draw Bresenham line. }
  WriteLong(REGS+DST_Y_X,longint(y1) shl 16+x1);

 { Allow setting of last pel bit and polygon outline bit for line drawing. }
  WriteLong(REGS+DP_CNTL_XDIR_YDIR_YMAJOR,flags);
  WriteLong(REGS+DST_BRES_ERR, err);
  WriteLong(REGS+DST_BRES_INC, add);
  WriteLong(REGS+DST_BRES_DEC, sub);
  WriteLong(REGS+DST_BRES_LNTH,(1-err)); { should be computed by ATI :) }

  WriteLong(REGS+DP_CNTL, save_dp_cntl);
  WriteLong(REGS+DP_DATATYPE, save_dp_datatype);

 end;


Procedure DrawBitmap(X,Y:integer; Bitmap:Array of Longint; W,H,ForeColor,BackColor:integer);
 var
  i:integer;
 begin
  WaitForFifo(5);
  WriteLong(REGS+DP_GUI_MASTER_CNTL,
   (0 shl 0) or      { Use mmDEFAULT_OFFSET and mmDEFAULT_PITCH for SRC }
   (0 shl 1) or      { Use mmDEFAULT_OFFSET and mmDEFAULT_PITCH for DST }
   (0 shl 2) or      { Use mmDEFAULT_SC_BOTTOM_RIGHT                    }
   (1 shl 3) or      { Use mmSC_TOP_LEFT and mmSC_BOTTOM_RIGHT for DST  }
   ($C shl 4) or     { NO_BRUSH needed.                                 }
   (BPPi shl 8) or   { Set DST_DATATYPE to current bpp                  }
   (0 shl 12) or     { Expand to foreground and background.             }
   (1 shl 14) or     { Consume monochrome data from LSbit to MSbit      }
   (0 shl 15) or     { Set Conversion temp to 6500k                     }
   ($CC shl 16) or   { Set ROP3 to SRC_COPY                             }
   (3 shl 24) or     { Source Data is from HOSTDATA registers.          }
   (0 shl 27) or     { Clear 3D_SCALE_CNTL (Disable 3D engine)          }
   (1 shl 28) or     { Clear CLR_CMP_CNTL (Disable Color Compare)       }
   (1 shl 29) or     { Clear AUX_SC_CNTL (Disable Auxilary Scissors)    }
   (1 shl 30) or     { Set DP_WRITE_MASK to $FFFFFFFF                   }
   (0 shl 31)        { No BRUSH_X_Y required.                           }
  );
 { Set the colors to expand the data to. }
  WriteLong(REGS+DP_SRC_FRGD_CLR,GetColor(ForeColor));
  WriteLong(REGS+DP_SRC_BKGD_CLR,GetColor(BackColor));
 { Setup the destination trajectory. }
  WriteLong(REGS+DST_X_Y,longint(x) shl 16+y);
  WriteLong(REGS+DST_WIDTH_HEIGHT, longint(h) shl 16+w);
 { Write the data out to the HostData registers.  We write the number of DWords
   less the last one, which we must write out to HOST_DATA_LAST to tell the GUI
   engine that the HOSTDATA operation is complete. }
  for i:=Low(Bitmap) to High(bitmap)-1 do begin
   WaitForFifo(1);
   WriteLong(REGS+HOST_DATA0,Bitmap[i]);
  end;
 { Write out the final piece of data. }
  WaitForFifo(1);
  WriteLong(REGS+HOST_DATA_LAST,Bitmap[High(Bitmap)]);

 end;

Procedure SetBrush(Pattern:array of Longint);
 var
  i:integer;
 begin
 { Set the drawing controls registers. *** missing in ATI demo :) *** }
  WaitForFifo(1);
  WriteLong(REGS+DP_MIX, ROP3_PATCOPY);
  for i:=Low(Pattern) to High(Pattern) do begin
   WaitForFifo(1);
   WriteLong(REGS+BRUSH_DATA0+4*i,Pattern[i]);
  end;
 end;

Procedure FillBrush(x,y,w,h,ForeColor,BackColor:integer);
 begin
 { We will use a 32x32 mono pattern, expanding it to the foreground,
   and background. }
  WaitForFifo(1);
  WriteLong(REGS+DP_DATATYPE, SRC_DSTCOLOR or BRUSH_32x32MONO or BPPi);
 { draw }
  WaitForFifo(5);
  WriteLong(REGS+DP_BRUSH_FRGD_CLR,GetColor(ForeColor));
  WriteLong(REGS+DP_BRUSH_BKGD_CLR,Getcolor(BackColor));
  WriteLong(REGS+BRUSH_Y_X,(32-y mod 32) shl 8 + (32-x mod 32));
  WriteLong(REGS+DST_Y_X,longint(y) shl 16 + x);
  WriteLong(REGS+DST_HEIGHT_WIDTH,longint(h) shl 16+w);
 end;

Procedure BrushLine(x1,y1,x2,y2,BrushType,ForeColor,BackColor:integer);
 var
  dx,dy:integer;
  flags:longint;
  err,add,sub:longint;
  save_dp_cntl,save_dp_datatype:longint;
 begin
  flags:=0;
  dx:=x2-x1; if dx<0 then dx:=-dx else flags:=          1 shl 31;
  dy:=y2-y1; if dy<0 then dy:=-dy else flags:=flags or (1 shl 15);

 { Calculate Bresenham parameters and draw line. }
  if dx<dy then begin
   flags:=flags or (1 shl 2);
   err:=-dy;
   add:=2*dx;
  end else begin
   err:=-dx;
   add:=2*dy;
  end;

  sub:=2*err; { should be computed by ATI :) }

  WaitForFifo(11);

  save_dp_cntl     := ReadLong(REGS+DP_CNTL);
  save_dp_datatype := ReadLong(REGS+DP_DATATYPE);

 { Set DP_DATATYPE }
  WriteLong(REGS+DP_DATATYPE, (BPPi or (BrushType shl 8) or ROP3_PATCOPY));
  WriteLong(REGS+DP_BRUSH_FRGD_CLR, GetColor(Forecolor));
  WriteLong(REGS+DP_BRUSH_BKGD_CLR, GetColor(Backcolor));

 { Draw Bresenham line. }
  WriteLong(REGS+DST_Y_X,longint(y1) shl 16+x1);

 { Allow setting of last pel bit and polygon outline bit for line drawing. }
  WriteLong(REGS+DP_CNTL_XDIR_YDIR_YMAJOR,flags);
  WriteLong(REGS+DST_BRES_ERR, err);
  WriteLong(REGS+DST_BRES_INC, add);
  WriteLong(REGS+DST_BRES_DEC, sub);
  WriteLong(REGS+DST_BRES_LNTH,(1-err)); { should be computed by ATI :) }

  WriteLong(REGS+DP_CNTL, save_dp_cntl);
  WriteLong(REGS+DP_DATATYPE, save_dp_datatype);

 end;

procedure StrechBitBlt(sx,sy,sw,sh,dx,dy,dw,dh:integer);
 var
  save_tex_cntl,
  save_scale_3d_cntl,
  save_dp_datatype,
  save_dp_mix:longint;
 begin
 { Save the control registers that will be modified. }
  WaitForFifo(4);
  save_tex_cntl     :=ReadLong(REGS+TEX_CNTL);
  save_scale_3d_cntl:=ReadLong(REGS+SCALE_3D_CNTL);
  save_dp_datatype  :=ReadLong(REGS+DP_DATATYPE);
  save_dp_mix       :=ReadLong(REGS+DP_MIX);

 { Set the drawing direction to left->right, top->bottom }
  WaitForFifo(1);
  WriteLong(REGS+DP_CNTL, DST_X_LEFT_TO_RIGHT or DST_Y_TOP_TO_BOTTOM);

  WaitForFifo(18);

 { Set the mix to a source copy, with a rectangular source }
  WriteLong(REGS+DP_MIX, ROP3_SRCCOPY or DP_SRC_RECT);

 { set this to NO BRUSH! }
  WriteLong(REGS+DP_DATATYPE, BPPi or (15 shl 8) or SRC_DSTCOLOR);

 { enable scaling with blending }
  WriteLong(REGS+SCALE_3D_CNTL, $00000040);

 { disable any texture features }
  WriteLong(REGS+TEX_CNTL, $00000000);
  WriteLong(REGS+MISC_3D_STATE_CNTL_REG, $00000100);

 { Set up the front end scaler }
  WriteLong(REGS+SCALE_3D_DATATYPE,BPPi);
  WriteLong(REGS+CLR_CMP_MASK_3D, $FFFFFFFF);
  WriteLong(REGS+SCALE_SRC_HEIGHT_WIDTH,longint(sh) shl 16+sw);

 { calculate the offset of the scaler data }
  WriteLong(REGS+SCALE_OFFSET_0,Width*sy*BPC+sx);

 { set SCALE_PITCH
   Because we've loaded the image into offscreen memory in a rectangular
   trajectory, our scaler pitch will be equal to the screen pitch. }
  WriteLong(REGS+SCALE_PITCH, Pitch{R128_AdapterInfo.pitch});

 { Disable any motion compensation features }
  WriteLong(REGS+MC_SRC1_CNTL, $00000000);

  WriteLong(REGS+SCALE_X_INC,Round(65535*sw/dw));
  WriteLong(REGS+SCALE_Y_INC,Round(65535*sh/dh));

  WriteLong(REGS+SCALE_HACC, $00000000);
  WriteLong(REGS+SCALE_VACC, $00000000);
  WriteLong(REGS+SRC_Y_X, longint(sy) shl 16+sx);
  WriteLong(REGS+SCALE_DST_X_Y, longint(dx) shl 16+dy);
  WriteLong(REGS+SCALE_DST_HEIGHT_WIDTH, longint(dh) shl 16+dw);

 { Now restore the registers we changed. }
  WaitForFifo (4);
  WriteLong(REGS+TEX_CNTL, save_tex_cntl);
  WriteLong(REGS+SCALE_3D_CNTL, save_scale_3d_cntl);
  WriteLong(REGS+DP_DATATYPE, save_dp_datatype);
  WriteLong(REGS+DP_MIX, save_dp_mix);

  WriteLong(REGS+MISC_3D_STATE_CNTL_REG, $0000000); (* not in the doc ! *)
 end;

Procedure TransBlit(x,y,w,h,dx,dy,mode,srccolor,dstcolor:integer);
 var
  save_dp_mix,
  save_dp_cntl,
  save_dp_datatype:longint;
 begin
  WaitForFifo (3);
  save_dp_mix     :=ReadLong(REGS+DP_MIX);
  save_dp_cntl    :=ReadLong(REGS+DP_CNTL);
  save_dp_datatype:=ReadLong(REGS+DP_DATATYPE);

  WaitForFifo (3);
 { set up the engine trajectory registers for a blt. }
  WriteLong(REGS+DP_MIX, ROP3_SRCCOPY or DP_SRC_RECT);
  WriteLong(REGS+DP_CNTL, DST_X_LEFT_TO_RIGHT or DST_Y_TOP_TO_BOTTOM);
  WriteLong(REGS+DP_DATATYPE, BPPi or SRC_DSTCOLOR or BRUSH_SOLIDCOLOR);

  WaitForFifo (4);
 { set up the transparency function in the colour compare circuitry }
  WriteLong(REGS+CLR_CMP_CLR_DST, GetColor(dstcolor));
  WriteLong(REGS+CLR_CMP_CLR_SRC, GetColor(srccolor));
  WriteLong(REGS+CLR_CMP_MASK,$FFFFFFFF);
  WriteLong(REGS+CLR_CMP_CNTL,
  {longint(cmpmode) shl 24+(dstmode shl  8)+(srcmode) }
   longint(mode and $300) shl (24-8) { Compare Mode = None, Src, Dst, Both }
         +(mode and $00F) shl 8      { Destination  = False, True, = , <> }
         +(mode and $0F0) shr 4      { Source Color = False, True, = , <> , NOT }
  );

 { Set up source and destination x and y values }
  WaitForFifo(3);
  WriteLong(REGS+SRC_Y_X, longint(y) shl 16+x);
  WriteLong(REGS+DST_Y_X,longint(dy) shl 16+dx);
  WriteLong(REGS+DST_HEIGHT_WIDTH,longint(h) shl 16+w);

 { Restore the modified registers }
  WaitForFifo(3);
  WriteLong(REGS+DP_MIX, save_dp_mix);
  WriteLong(REGS+DP_CNTL, save_dp_cntl);
  WriteLong(REGS+DP_DATATYPE, save_dp_datatype);
 end;

{ Microcode for the CCE Microengine (256 QWORDS).  Magic values, do not touch! }
Const
 Microcode:array[0..511] of longint=({$I MICRO.INC});
(**
Procedure InitCEE;
 var
  i:integer;
 begin

    DWORD cce_buf_size, temp;

 { Load CCE Microcode }
  { Wait for engine idle before loading the microcode. }
    WaitForIdle;
  { Set starting address for writing the microcode. }
    WriteLong(REGS+PM4_MICROCODE_ADDR, 0);
    for i:=0 to 255 do begin
     { The microcode write address will automatically increment after
       every microcode data high/low pair.  Note that the high DWORD
       must be written first for address autoincrement to work correctly. }
     WriteLong(REGS+PM4_MICROCODE_DATAH,Microcode[2*i]);
     WriteLong(REGS+PM4_MICROCODE_DATAL,Microcode[2*i+1]);
    end;
 { ------------------ }

 { Perform a soft reset of the engine }
   ResetEngine;

    CCEFifoSize = CCEmode[index].fifosize;
    CCEBMFlag = CCEmode[index].busmaster;
    if (CCEBMFlag)
    {
        R128_CCESubmitPackets = CCESubmitPacketsBM;
        if (CCESetupBusMaster ())
        {
            return (CCE_FAIL_BUS_MASTER_INIT);
        } // if
        cce_buf_size = RING_SIZE_LOG2;
        bm_save_state = regr (BUS_CNTL);
        regw (BUS_CNTL, (bm_save_state & ~BUS_MASTER_DIS));
    }
    else
    {
        R128_CCESubmitPackets = CCESubmitPacketsPIO;
        cce_buf_size = 0;
    } // if

    // Set the Rage 128 to requested CCE mode.
    CCERequestedMode = CCEmode[index].pm4buffermode + cce_buf_size;
    regw (PM4_BUFFER_CNTL, CCERequestedMode);

    temp = regr (PM4_BUFFER_ADDR);

    // Set the CCE to free running mode

    regw (PM4_MICRO_CNTL, PM4_MICRO_FREERUN);

	// Set vertex buffer and indirect buffer flags, and create if necessary.

    CCEVBFlag = CCEmode[index].vertexbuffer;
    CCEIBFlag = CCEmode[index].indirectbuffer;

    CreateCCEBuffers ();

 end;

Procedure DoneCEE;
 begin
 end;

Procedure SendCEE(Buf:array of longint; count:integer);
 begin
 end;
**)

begin
 asm { Get PC Charset pointer }
 push bp
  mov ax,1130h
  mov bh,02h
  int 10h
  mov word(rom8x8+0),bp
  mov word(rom8x8+2),es
 pop  bp
 end;
end.